
# This file was generated by 'versioneer.py' (0.11) from
# revision-control system data, or from the parent directory name of an
# unpacked source archive. Distribution tarballs contain a pre-generated copy
# of this file.

version_version = '0.5.0-2-g1889870'
version_full = '18898707b721ede2a3d3e8ba5ed7fbd22e409389'
def get_versions(default={}, verbose=False):
    return {'version': version_version, 'full': version_full}

